import common
import random
import os
import numpy as np
import cv2
import tensorrt as trt

from PIL import Image

from timeit import default_timer as timer

class ModelData(object):
    MODEL_PATH = "mobilenetV2.onnx"
    INPUT_SHAPE = (3, 224, 224)
    # We can convert TensorRT data types to numpy types with trt.nptype()
    DTYPE = trt.float32

# You can set the logger severity higher to suppress messages (or lower to display more messages).
TRT_LOGGER = trt.Logger(trt.Logger.VERBOSE)
trt_runtime = trt.Runtime(TRT_LOGGER)


# DESERIALIZAR/CARGAR EL MOTOR DE INFERENCIA
def load_engine(trt_runtime, plan_path):
    with open(plan_path, 'rb') as f:
        engine_data = f.read()
        engine = trt_runtime.deserialize_cuda_engine(engine_data)
    return engine


def preprocess(img_path, img_size):
    img = cv2.imread(img_path, cv2.IMREAD_COLOR)[:, :, ::-1]
    img = np.float32(img)/255.0
    img = cv2.resize(img, (img_size, img_size))
    return img

def main():
    files = os.listdir("./data/mobilenetV2")
    data_files = []
    for fil in files:
        fil_split = fil.split('.')
        extension = fil_split[1]
        if extension == 'jpg':
            data_files.append(fil)

    data_files.append(ModelData.MODEL_PATH)
    data_files.append('class_labels.txt')

    # Set the data path to the directory that contains the trained models and test images for inference.
    _, data_files = common.find_sample_data(description="Runs a MobilenetV2 network with a TensorRT inference engine.", subfolder="mobilenetV2",find_files= data_files)

    # Get test images, models and labels.
    test_images = data_files[0:len(data_files)-2]
    onnx_model_file, labels_file = data_files[len(data_files)-2:]
    labels = open(labels_file, 'r').read().split('\n')

    print('Creating context...')
    context_timer_start = timer()
    model_engine = load_engine(trt_runtime, "./mobilenetV2_fp32.plan")

    with model_engine as engine:
        # Inference is the same regardless of which parser is used to build the engine, since the model architecture is the same.
        # Allocate buffers and create a CUDA stream.
        inputs, outputs, bindings, stream = common.allocate_buffers(engine)
        # Contexts are used to perform inference.
        with engine.create_execution_context() as context:
            context_timer_end = timer()
            print('Context created in {} seconds'.format(context_timer_end-context_timer_start))
            i=0
            pre_process_times=np.zeros(len(test_images))
            inference_times=np.zeros(len(test_images))
            image_process_loop_start = timer()
            for test_image in test_images:
                pre_process_start = timer()
                # Load a normalized test case into the host input page-locked buffer.
                #test_image = random.choice(test_images)
                img = preprocess(test_image,224)
                inputs[0].host = img
                pre_process_end = timer()
                pre_process_time = pre_process_end-pre_process_start
                pre_process_times[i]=pre_process_time
                print('Image preprocessed in {} seconds'.format(pre_process_time))

                # Run the engine. The output will be a 1D tensor of length 1000, where each value represents the
                # probability that the image corresponds to that label
                inference_start = timer()
                trt_outputs = common.do_inference_v2(context, bindings=bindings, inputs=inputs, outputs=outputs, stream=stream)
                inference_end = timer()
                inference_time = inference_end-inference_start
                if i == 0:
                    pass
                else:
                    inference_times[i-1]=inference_time
                print('Image inference took in {} seconds'.format(inference_time))

                trt_output_with_labels = zip(trt_outputs[0],labels)
                sorted_pairs = sorted(trt_output_with_labels,reverse=True)
                sorted_trt_output_with_labels = zip(*sorted_pairs)
                list_probability, list_labels = [ list(tuple) for tuple in  sorted_trt_output_with_labels]
                print('Test image {}:{}'.format(i,test_image))
                print('---------------------------------------------------')
                for j in range(5):
                    print("{}.{}:{}".format(j+1,list_labels[j],list_probability[j]))
                print('---------------------------------------------------')
                i=i+1
            image_process_loop_end = timer()
            image_process_loop_time = image_process_loop_end - image_process_loop_start

            print('{} images were processed in {} seconds'.format(i,image_process_loop_time))
            print('Average time for image preprocessing is {} seconds'.format(np.average(pre_process_times)))
            print('Average time for image inferece is {} seconds'.format(np.average(inference_times)))

if __name__ == '__main__':
    main()
